# from .Phase import Phase
from typing import Dict
from pyPhases.storage.Storage import Storage
from pyPhases.Data import Data, DataNotFound
from pyPhases.Stage import Stage
from pyPhases.decorator.ExportValidator import ExportValidator
from pyPhases.util.Logger import classLogger, Logger, LogLevel
from pyPhases.util.Optionizable import Optionizable

class ConfigNotFoundException(Exception):
    pass

@classLogger
class Project(Optionizable):
    """
    Represents a whole project with several phases

    Parameters
    ----------
    name : string
        name of the project
    namespace : string
        namespace of the project e.e tud.ibmt
    dataStorage : storage
        the storage engine that should be used for storing all kinds of data
        (the default is a Filestorage pointing to the data/ directory)
    """

    phases = {}
    name = "myProject"
    namespace = ""
    stages = []
    classes = []
    exporters = []
    classesMap = {}
    registeredData = {}
    config = {}
    dataStorage = []
    stageIndex = 0
    phaseIndex = 0
    decorators = []
    debug = False
    testRun = False
    phaseMap = {}

    def __init__(self):
        self.registerDecorator(ExportValidator())

    def addStorage(self, storage):
        self.dataStorage.append(storage)

    def registerDecorator(self, decorator):
        self.logDebug("Register Decorator: " + type(decorator).__name__)
        self.decorators.append(decorator)

    def registerPublisher(self, publisher):
        self.logDebug("Register Publisher: " + type(publisher).__name__)
        self.registerDecorator(publisher)

    def registerExporter(self, exporter):
        self.logDebug("Register Exporter: " + type(exporter).__name__)
        self.exporters.append(exporter)

    def setClasses(self, classes):
        self.classes = classes
        for index, className in enumerate(self.classes):
            self.classesMap[className] = index

    def addStage(self, stageName: str, autorun=True):
        """ creates a stage for the project, a stage is a sequential list of phases
        Parameters
        ----------
        stageName : str
            a unique name of the stage that should be created
        """
        stage = Stage(stageName, {
            "autorun": autorun
        })
        self.stages += [stage]
        self.phases[stageName] = []

    def getExporterForIntsance(self, instance):
        self.logDebug("Get Exporter For: " + type(instance).__name__)
        return self.getExporterForType(type(instance))

    def getExporterForType(self, theType, storage: Storage = None):
        exporters = self.exporters

        if storage != None and storage.acceptedExporter != None:
            exporters = storage.acceptedExporter

        for exporter in exporters:
            self.logDebug("Check: " + type(exporter).__name__)
            if exporter.checkType(theType):
                self.logDebug("Found exporter")
                return exporter.reinit()
        return None

    def registerStream(self, dataName, expectedReturnType, options, version: str = "current"):

        dataObj = Data.getFromName(dataName)
        dataObj.version = version
        dataObj.setProject(self)
        dataId = dataObj.getDataId()

        try:
            exporter = self.getExporterForType(expectedReturnType)
            return exporter.stream(dataId, options)
        except Exception:
            raise Exception("No stream exporter found for this")

    def dataExists(self, data: Data):
        return data.getDataId() in self.registeredData

    def getData(self,
                dataName: str,
                expectedReturnType=None,
                version: str = "current",
                options={}):

        dataObj = Data.getFromName(dataName)
        dataObj.version = version
        dataObj.setProject(self)
        dataId = dataObj.getDataId()

        self.logDebug("Try to get Data: " + dataId)

        # just generated Data
        if dataId in self.registeredData:
            self.logDebug("Data in memory: " + dataId)
            return self.registeredData[dataId]

        # load from storage layer
        try:
            for storage in self.dataStorage:
                self.logDebug("Check storage: " + type(storage).__name__)
                exporter = self.getExporterForType(expectedReturnType, storage)
                if exporter is None:
                    raise Exception("No exporter found for this type: %s" % (expectedReturnType))

                if exporter.includesStorage:
                    return exporter.importData(dataId, options)

                dataBytes = storage.read(dataId)
                self.logDebug("Data in storage: " + type(storage).__name__)
                return exporter.importData(dataBytes, options)

            exporter = self.getExporterForType(expectedReturnType)
            if exporter != None and exporter.includesStorage:
                return exporter.importData(dataId, options)

        except DataNotFound:
            pass

        # regenerate from previous stage/s
        if (version == "current"):
            self.logWarning("Data " + dataId +
                            " was not found, rerunning previous stages for current data")

            for phase in self.getPhases():
                if dataName in phase.exportDataStrings:
                    phase.getData(dataName)
                    return self.getData(dataName, expectedReturnType, version, options)

        raise DataNotFound("Data " + dataName + " was not found")

    def registerData(self,
                     dataName: str,
                     data: str,
                     version: str = "current",
                     save: bool = True):

        dataObj = Data.getFromName(dataName)
        dataObj.version = version
        dataObj.setProject(self)
        dataId = dataObj.getDataId()

        # save to runtime project
        self.registeredData[dataId] = data

        if (save == False):
            return

        if (self.dataStorage == None):
            self.logWarning("There was no datastorage registerd")
            return

        # save to storage layer
        exporter = self.getExporterForIntsance(data)

        if (exporter == None):
            self.logWarning("No exporter for datatype (" + type(data).__name__ +
                            ") the data " + dataName + " will not be automaticly save")
            return

        if exporter.includesStorage:
            exporter.exportDataId(dataId, data)
        else:
            dataBytes = exporter.export(data, dataId)
            for storage in self.dataStorage:
                storage.write(dataId, dataBytes)

    def setConfig(self, name: str, value):
        self.config[name] = value

    def getConfig(self, name: str) -> str:
        if not name in self.config:
            raise ConfigNotFoundException("The Config entry '%s' was not found, if you rely on a earlier config entry in a previous phase make sure you put the config entry in the __init__ section of the phase" % (name))
        return self.config[name]

    def addConfig(self, config) -> None:
        for name in config:
            if (name in self.config):
                Warning("The config name " + name + " was specified in multiple phases and was overwritten by the latest!")

            self.config[name] = config[name]

    def addPhase(self, phase, stage: str = "", name: str = None):
        if (name != None):
            phase.name = name

        self.addConfig(phase.config)
        phase.project = self

        if (not stage in self.phases):
            raise Exception('The stage to this phase was not added before!')

        self.phaseMap[type(phase).__name__] = phase
        self.phases[stage] += [phase]
        phase.prepare()

    def runAllStages(self):
        for stage in self.stages:
            if stage.getOption("autorun"):
                self.runStage(stage.name)

    def runStage(self, stageName):
        self.log("RUN stage " + stageName)
        for phase in self.phases[stageName]:
            phase.run()

    def run(self, stageName=None):
        if (stageName == None):
            self.runAllStages()
        else:
            self.runStage(stageName)

    def getPhase(self, name: str):
        return self.phaseMap[name]

    def getPhases(self):
        for stage in self.stages:
            for phase in self.phases[stage.name]:
                yield phase
